/*
 * bonobo-storage-efs.c: libefs based Storage implementation
 *
 * Author:
 *   Dietmar Maurer (dietmar@maurer-it.com)
 *
 *
 * Problems:
 *
 * do we really need open and create functions? maybe we can use open and
 * a special flag O_CREATE (like in stdio)
 *
 * storage_copy_to: ??? 
 *      idl: copy_to (in string path, in Storage dest, is string new_path)
 *
 * list_contents: bad interface! redesign necessary?
 *
 */

#include <config.h>
#include <efs.h>
#include <storage-modules/bonobo-storage-efs.h>
#include <storage-modules/bonobo-stream-efs.h>
#include <bonobo/bonobo-storage-plugin.h>

/*
 * Creates and activates the corba server 
 */
static BonoboStorage *
create_storage_efs_server (const BonoboStorageEFS *storage_efs)
{
	Bonobo_Storage corba_storage;

	corba_storage = bonobo_storage_corba_object_create (
		BONOBO_OBJECT (storage_efs));
				 
	return bonobo_storage_construct (BONOBO_STORAGE (storage_efs), 
					 corba_storage);
}

static void
bonobo_storage_efs_destroy (GtkObject *object)
{
	BonoboStorageEFS *storage_efs = BONOBO_STORAGE_EFS (object);
	
	if (storage_efs->owner) {
		if (storage_efs->dir) efs_dir_close (storage_efs->dir);
		bonobo_object_unref (BONOBO_OBJECT (storage_efs->owner));    
	} else {
		if (storage_efs->dir)
			efs_close (storage_efs->dir);
	}
}

static Bonobo_StorageInfo*
real_get_info (BonoboStorage *storage,
	       const CORBA_char *path,
	       const Bonobo_StorageInfoFields mask,
	       CORBA_Environment *ev)
{
	g_warning ("Not implemented");

	return CORBA_OBJECT_NIL;
}

static void
real_set_info (BonoboStorage *storage,
	       const CORBA_char *path,
	       const Bonobo_StorageInfo *info,
	       const Bonobo_StorageInfoFields mask,
	       CORBA_Environment *ev)
{
	g_warning ("Not implemented");
}

static BonoboStream *
real_open_stream (BonoboStorage *storage, const CORBA_char *path, 
		  Bonobo_Storage_OpenMode mode, CORBA_Environment *ev)
{
	BonoboStorageEFS *storage_efs = BONOBO_STORAGE_EFS (storage);

	return bonobo_stream_efs_open (storage_efs, path, mode);
}


static BonoboStorage *
real_open_storage (BonoboStorage *storage, const CORBA_char *path, 
		   Bonobo_Storage_OpenMode mode, CORBA_Environment *ev)
{
	BonoboStorageEFS *storage_efs = BONOBO_STORAGE_EFS (storage);
	BonoboStorageEFS *sefs;
	EFSDir *dir;
	gint efs_mode;

	efs_mode = bonobo_mode_to_efs (mode);

	if (efs_dir_open (&dir, storage_efs->dir, path, efs_mode)) {
		CORBA_exception_set (ev, CORBA_USER_EXCEPTION, 
				     ex_Bonobo_Storage_NotFound, NULL);
		return NULL;
	}

	sefs = gtk_type_new (bonobo_storage_efs_get_type ());
	sefs->dir = dir;
	sefs->owner = storage;
	bonobo_object_ref (BONOBO_OBJECT (storage));
	create_storage_efs_server (sefs);

	return BONOBO_STORAGE (sefs);
}

static void
real_copy_to (BonoboStorage *storage, Bonobo_Storage target, 
	      CORBA_Environment *ev)
{	
	g_warning ("Not yet implemented");
}

static void
real_rename (BonoboStorage *storage, const CORBA_char *path, 
	     const CORBA_char *new_path, CORBA_Environment *ev)
{
	BonoboStorageEFS *storage_efs = BONOBO_STORAGE_EFS (storage);

	if (efs_rename (storage_efs->dir, path, new_path)) {
		CORBA_exception_set (ev, CORBA_USER_EXCEPTION, 
				     ex_Bonobo_Storage_NotFound, NULL);
	}
}

static void
real_commit (BonoboStorage *storage, CORBA_Environment *ev)
{
	BonoboStorageEFS *storage_efs = BONOBO_STORAGE_EFS (storage);

	if (storage_efs->owner)
		return;

	efs_commit (storage_efs->dir);
}

static void
real_revert (BonoboStorage *storage, CORBA_Environment *ev)
{
	BonoboStorageEFS *storage_efs = BONOBO_STORAGE_EFS (storage);

	if (storage_efs->owner)
		return;

	efs_revert (storage_efs->dir);
}


static Bonobo_Storage_DirectoryList *
real_list_contents (BonoboStorage *storage, const CORBA_char *path, 
		    Bonobo_StorageInfoFields mask, CORBA_Environment *ev)
{
	BonoboStorageEFS *storage_efs = BONOBO_STORAGE_EFS (storage);
	Bonobo_Storage_DirectoryList *list;
	EFSDir *dir;
	int i;

	list = Bonobo_Storage_DirectoryList__alloc();
	CORBA_sequence_set_release (list, TRUE); 

	if (!efs_dir_open (&dir, storage_efs->dir, path, 0)) {
		EFSDirEntry de;

		for (i = 0; !(efs_dir_read (dir, &de)); i++)
			;
		efs_dir_close (dir);
	} else
		goto list_contents_except;
		
	if (!efs_dir_open (&dir, storage_efs->dir, path, 0)) {
		EFSDirEntry de;
		Bonobo_StorageInfo *buf;
		
		buf = CORBA_sequence_Bonobo_StorageInfo_allocbuf (i);
	
		for (i = 0; !(efs_dir_read (dir, &de)); i++) {
			buf[i].name = CORBA_string_dup (de.name);
			if (de.type == EFS_DIR)
				buf[i].type = Bonobo_STORAGE_TYPE_DIRECTORY;
			else buf[i].type = Bonobo_STORAGE_TYPE_REGULAR;

			buf[i].content_type = CORBA_string_dup("");
			buf[i].size = 0;
		}

		list->_length = i;
		list->_buffer = buf;
		efs_dir_close (dir);

		return list;
	}

 list_contents_except:
	CORBA_exception_set (ev, CORBA_USER_EXCEPTION, 
			     ex_Bonobo_Storage_NotFound, NULL);
	CORBA_free (list);
	return NULL;
}


static void
real_erase (BonoboStorage *storage, const CORBA_char *path,
	    CORBA_Environment *ev)
{
	BonoboStorageEFS *storage_efs = BONOBO_STORAGE_EFS (storage);

	if (efs_erase (storage_efs->dir, path)) {
		CORBA_exception_set (ev, CORBA_USER_EXCEPTION, 
				     ex_Bonobo_Storage_NotFound, NULL);
	}
}

static void
bonobo_storage_efs_class_init (BonoboStorageEFSClass *class)
{
	GtkObjectClass     *oclass = (GtkObjectClass *) class;
	BonoboStorageClass *sclass = (BonoboStorageClass *) class;

	sclass->get_info       = real_get_info;
	sclass->set_info       = real_set_info;
	sclass->open_stream    = real_open_stream;
	sclass->open_storage   = real_open_storage;
	sclass->copy_to        = real_copy_to;
	sclass->rename         = real_rename;
	sclass->commit         = real_commit;
	sclass->revert         = real_revert;
	sclass->list_contents  = real_list_contents;
	sclass->erase          = real_erase;
	
	oclass->destroy = bonobo_storage_efs_destroy;
}

GtkType
bonobo_storage_efs_get_type (void)
{
	static GtkType type = 0;
  
	if (!type) {
		GtkTypeInfo info = {
			"IDL:GNOME/StorageEFS:1.0",
			sizeof (BonoboStorageEFS),
			sizeof (BonoboStorageEFSClass),
			(GtkClassInitFunc) bonobo_storage_efs_class_init,
			(GtkObjectInitFunc) NULL,
			NULL, /* reserved 1 */
			NULL, /* reserved 2 */
			(GtkClassInitFunc) NULL
		};

		type = gtk_type_unique (bonobo_storage_get_type (), &info);
	}

	return type;
}

/** 
 * bonobo_storage_efs_open:
 * @path: path to a file that represents the storage
 * @flags: flags
 * @mode: mode
 *
 * Opens a BonoboStorage object in @path. The storage opened is an activated
 * CORBA server for this storage.
 *
 * Returns the BonoboStorage GTK object.
 */
BonoboStorage *
bonobo_storage_efs_open (const gchar *path, gint flags, gint mode)
{
	BonoboStorageEFS *sefs;
	gint efs_flags = 0;

	efs_flags = bonobo_mode_to_efs (flags);

	sefs = gtk_type_new (bonobo_storage_efs_get_type ());

	if (efs_open (&sefs->dir, path, efs_flags, mode, NULL)) {
		bonobo_object_unref (BONOBO_OBJECT (sefs));
		return NULL;
	}
    
	if (!create_storage_efs_server (sefs)) {
		bonobo_object_unref (BONOBO_OBJECT (sefs));
		return NULL;
	}

	return BONOBO_STORAGE (sefs);
}

gint 
init_storage_plugin (StoragePlugin *plugin)
{
	g_return_val_if_fail (plugin != NULL, -1);

	plugin->name = "efs";
	plugin->description = "Embedded Filesystem Driver";
	plugin->version = VERSION;
	
	plugin->storage_open = bonobo_storage_efs_open; 

	return 0;
}
